;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; This program is a floppy disk bootsector. It's purpose is to dump the
; contents of the real-mode interrupt vector table (IVT; the first 1024
; bytes of memory) to a file on the floppy disk named "IVT.BIN"
;
; Chris Giese	<geezer@execpc.com>	http://my.execpc.com/~geezer/
; This code is public domain (no copyright).
; You can do whatever you want with it.
;
; 4 Jul 2015: initial release
;
; 1. Assemble this file with NASM:
;	nasm -f bin -o dump-ivt.bin dump-ivt.asm
;
; 2. Install the assembled version of this file as the bootsector
;    of a 1.44 Mbyte floppy:
;	(DOS)	partcopy dump-ivt.bin 0 200 -f0
;    -OR-
;	(DOS)	debug dump-ivt.bin
;		-w 100 0 0 1	    ('-' is DEBUG prompt. Type ? for help.)
;		-q
;    -OR-
;	(Linux)	cat dump-ivt.bin >/dev/fd0
;
; 3. Boot the PC from the floppy disk. ALL DATA ON THE FLOPPY DISK
;    WILL BE ERASED; and a 1024-byte file named "ivt.bin" will be
;    created; containing the IVT dump.
;
; Layout of 1.44 Mbyte floppy:
; sect.	count	description
;    0	   1	boot/reserved sector
;    1 	9 * 2	two FATs
;   19	  14	root directory
;   33	2847	data area
; 2880
;
; Full-blown FAT12 filesystem code won't fit into 512 bytes so I just
; erase the floppy, create FAT and root directory entries for a 1024-byte
; file, and write the contents of the file (the IVT) to sectors 33 and 34;
; which is where data starts on a normal, freshly-formatted FAT12 floppy.
;
; To erase the floppy:
; (1) - Write 0 bytes to sectors 1-32 (512 bytes per sector; all zeroed)
;     - Write these three hex values to byte offset 0 of sectors 1 and 10:
;	F0h FFh FFh
;
; To allocate 1K (clusters 2 and 3) of uninitialized disk space:
;     - Write these three hex values to byte offset 3 of sectors 1 and 10:
;	03h F0h FFh
;
; (2) - The allocation can be combined with the second step of the erase
;       operation by writing these six hex values to byte offset 0 of
;       sectors 1 and 10:
;	F0h FFh FFh 03h F0h FFh
;
; To create a 1024-byte file on the floppy named "IVT.BIN" and associate
; it with the previously-allocated disk space:
; (3) - Write these 32 hex values to byte offset 0 of sector 19:
;	49h 56h 54h 20h 20h 20h 20h 20h ; "IVT     "
;	42h 49h 4Eh			; "BIN"
;	20h		; DOS file attribute byte (b5=Archive)
;	00h		; reserved
;	00h		; 10 ms increments past ctime   (VFAT only)
;	00h 00h		; DOS-format file creation time (VFAT only)
;	00h 00h		; DOS-format file creation date (VFAT only)
;	00h 00h		; DOS-format last-accessed date (VFAT only)
;	00h 00h		; bits 31-16 of SCN (FAT32 only)
;	00h 00h		; DOS-format last-modified time (=midnight)
;	21h 00h		; DOS-format last-modified date (=Jan 1 1980)
;	02h 00h		; SCN=starting cluster number
;	00h 04h 00h 00h	; file size, in bytes
;
; To write the IVT to file "IVT.BIN":
; (4) - Write the 512 bytes at address 0000h:0000h to sector 33, then
;       write the 512 bytes at address 0000h:0200h to sector 34
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; use byte-offset addressing from BP for smaller code
%define	VAR(x)	((x) - start) + bp

; INT 0x13 drive number for the floppy = 0:
%define	DRIVE_NUM	0
%define	ADDR_STACK	7C00h
%define FS_ORG		100h

	CPU 8086

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; VARIABLES
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	ORG FS_ORG
start:

; (2 bytes) number of 16-byte paragraphs per sector
para_per_sector		EQU (start - 2)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; MINIMAL BIOS PARAMETER BLOCK (BPB)
; Values used here are for standard 1.44 meg floppy
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	jmp short skip_bpb
	nop
oem_id:			; offset 0x03 ( 3)
	db "_GEEZER_"	; Win 9x will scribble over the OEM ID
bytes_per_sector:	; offset 0x0B (11)
	dw 512
sectors_per_cluster:	; offset 0x0D (13)
	db 1
fat_start:
num_reserved_sectors:	; offset 0x0E (14)
	dw 1
num_fats:		; offset 0x10 (16)
	db 2
num_root_dir_ents:	; offset 0x11 (17)
	dw 224		; =7168 bytes =14 sectors
total_sectors:		; offset 0x13 (19)
	dw 18 * 2 * 80
media_id:		; offset 0x15 (21)
	db 0xF0
sectors_per_fat:	; offset 0x16 (22)
	dw 9
sectors_per_track:	; offset 0x18 (24)
	dw 18
heads:			; offset 0x1A (26)
	dw 2
hidden_sectors:		; offset 0x1C (28)
	dd 0
total_sectors_large:	; offset 0x20 (32)
	dd 0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; CODE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

skip_bpb:
%ifndef DOS
	jmp ((7C00h - FS_ORG) / 16):fix_cs
fix_cs:
	mov ax,cs
	mov ds,ax
	mov es,ax
	mov ss,ax
	mov sp,ADDR_STACK
%endif
	cld
	mov bp,start

; calculate the number of 16-byte paragraphs per sector
	mov ax,[VAR(bytes_per_sector)]
	mov cl,4
	shr ax,cl
	mov [VAR(para_per_sector)],ax

; say hello
	mov si,_start_msg
	call puts

; (step 1)
; zero sector buffer
	mov di,_buf
	xor ax,ax
	mov cx,256
	rep stosw

; zero sectors 1-32
	xor dx,dx
	mov ax,1
	mov bx,_buf
	mov cx,32
	call write_sectors_chs
	jc error

; (step 2)
; finish initializing FATs and immediately allocate clusters 2 and 3
	mov di,_buf
	mov si,_fat_magic
	mov cx,fat_magic_len
	rep movsb

;	xor dx,dx
;	mov ax,1		; first FAT
;	mov bx,_buf
	mov cx,1
	call write_sectors_chs
	jc error

	mov ax,10		; second FAT
	call write_sectors_chs
	jc error

; (step 3)
; allocate directory entry
	mov di,_buf
	mov si,_dir_magic
	mov cx,dir_magic_len
	rep movsb

;	xor dx,dx
	mov ax,19		; root directory starts here
;	mov bx,_buf
	mov cx,1
	call write_sectors_chs
	jc error

; (step 4)
; write IVT to sectors 33 and 34
;	xor dx,dx
	mov ax,33		; data area of floppy starts here
	xor bx,bx
	mov es,bx
;	mov cx,1
	call write_sectors_chs
	jc error

	inc ax
	mov bx,0x0200
	call write_sectors_chs
	jc error

	mov si,_ok_msg
	jmp short do_msg
error:
	mov si,_error_msg
do_msg:
	call puts

%ifdef DOS
; exit to DOS
	mov ax,0x4C00
	int 0x21
%else
	mov si,_reboot_msg
	call puts

; await key pressed
	mov ah,0
	int 0x16

; re-start the boot process
	int 0x19
%endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			write_sectors_chs
; action:		writes one or more disk sectors using INT 0x13 AH=0x03
; in:			DX:AX=LBA number of sector to write (relative to
;			start of partition), CX=sector count, ES:BX -> buffer
; out (disk error):	CY=1
; out (success):	CY=0
; modifies:		(nothing)
; minimum CPU:		8088
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

write_sectors_chs:
	push es
	push di
	push dx
	push cx
	push ax

; DX:AX=LBA sector number
; add partition start (= number of hidden sectors)
		add ax,[VAR(hidden_sectors + 0)]
		adc dx,[VAR(hidden_sectors + 2)]
		inc cx
		jmp short rsc_3
rsc_1:
		push dx
		push cx
		push ax

; DX:AX=LBA sector number
; divide by number of sectors per track to get sector number
; Use 32:16 DIV instead of 64:32 DIV for 8088 compatability
; Use two-step 32:16 divide to avoid overflow
			mov cx,ax
			mov ax,dx
			xor dx,dx
			div word [VAR(sectors_per_track)]
			xchg cx,ax
			div word [VAR(sectors_per_track)]
			xchg cx,dx

; DX:AX=quotient, CX=remainder=sector (S) - 1
; divide quotient by number of heads
			mov di,ax
			mov ax,dx
			xor dx,dx
			div word [VAR(heads)]
			xchg di,ax
			div word [VAR(heads)]
			xchg di,dx

; DX:AX=quotient=cylinder (C), DI=remainder=head (H)
; error if cylinder >=1024
			or dx,dx	; DX != 0; so cyl >= 65536
			stc
			jne rsc_2
			cmp ah,4	; AH >= 4; so cyl >= 1024
			cmc
			jb rsc_2

; move variables into registers for INT 0x13 AH=0x03
			mov dx,di
			mov dh,dl	; DH=head
			inc cx		; CL5:0=sector
			mov ch,al	; CH=cylinder 7:0
			shl cl,1
			shl cl,1
			shr ah,1
			rcr cl,1
			shr ah,1
			rcr cl,1	; CL7:6=cylinder 9:8
			mov dl,DRIVE_NUM

; INT 0x13 AH=0x03 called once for each sector. Multi-sector writes
; may fail if we cross a track or 64K boundary
			mov ax,0x0301	; AL=num_sectors
			int 0x13
			jnc rsc_2
; reset drive
			xor ax,ax
			int 0x13
			jc rsc_2
; try write a second time
			mov ax,0x0301
			int 0x13
rsc_2:
		pop ax
		pop cx
		pop dx
		jc rsc_4

; increment segment part of address and LBA sector number, and loop
		mov di,es
		add di,[VAR(para_per_sector)]
		mov es,di
		inc ax
		jne rsc_3
		inc dx
rsc_3:
		loop rsc_1
rsc_4:
	pop ax
	pop cx
	pop dx
	pop di
	pop es
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:			puts
; action:		displays text string
; in:			DS:SI -> 0-terminated string
; out:			(nothing)
; modifies:		AX, BX
; minimum CPU:		8088
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

puts:
	mov ah,0x0E		; INT 0x10: TELETYPE OUTPUT
	mov bh,0		; page 0
	jmp short puts_2
puts_1:
	int 0x10
puts_2:
	lodsb
	or al,al
	jne puts_1
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; DATA
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

_fat_magic:
	db 0xF0, 0xFF, 0xFF, 0x03, 0xF0, 0xFF
fat_magic_len equ ($ - _fat_magic)

_dir_magic:
	db "IVT     BIN", 0x20
	times 12 db 0
	dw 0x21
	dw 2
	dd 1024
dir_magic_len equ ($ - _dir_magic)

_start_msg:
	db "Erasing floppy and writing IVT to file 'ivt.bin'...", 13, 10, 0
_ok_msg:
	db "Success", 13, 10, 0
_error_msg:
	db "Error writing to floppy", 13, 10, 0
_reboot_msg:
	db "Eject floppy & press a key to reboot", 13, 10, 0

; pad with NOPs to offset 510
	times (510 + $$ - $) nop

; 2-byte magic bootsector signature
	db 0x55, 0xAA

; 512-byte sectors buffer at address 0x7E00:
_buf:
