;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Fast reboot from DOS
; Chris Giese	<geezer@execpc.com>	http://my.execpc.com/~geezer/
;
; This file is meant to be used with UNBOOTC.C; see notes in that file.
;
; Build with NASM: nasm [-g] -f obj unboota.asm
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

GROUP DGROUP _TEXT _DATA ; TINY memory model

SEGMENT _TEXT PUBLIC CLASS=CODE ALIGN=1

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:		unreal
; action:	enables (AX != 0) or disables (AX == 0) unreal mode
; inputs:	AX
; outputs:	interrupts disabled
; minimum CPU:	386+
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

unreal:
	push gs
	push fs
	push es
	push ds
	push dx
	push eax

; enable or disable unreal mode?
		or ax,ax
		je disable
		mov al,0xCF
disable:
		mov [_gdt1 + 6],al

; The GDT address in the GDT "pseudo-descriptor" is a LINEAR address;
; not subject to normal, segment-based addressing. It must be "patched"
; or "fixed up" by adding the segment base address, which is DS * 16:
		mov ax,ds
		mov dx,16
		mul dx

		add ax,_gdt
		adc dx,byte 0
		mov [_gdt_ptr + 2],ax
		mov [_gdt_ptr + 4],dx

; disable interrupts by clearing FLAGS.IF
; (and clear IOPL and NT bits in FLAGS)
		push dword 2
		popfd

; tell the CPU where the GDT is, and how big it is
		lgdt [_gdt_ptr]

; set the PE bit in register CR0 (=MSW)
		mov eax,cr0
		or al,1
		mov cr0,eax

; Load segment registers with descriptors for segments w/ limit
; 64K-1 (disable unreal mode) or 4GB-1 (enable unreal mode).
; The descriptors have base address 0, which is not what we want,
; but the base addresses will go back their old values when we pop
; the segment registers in real mode; below.
		mov dx,LINEAR_SEL
		mov ds,dx
		mov es,dx
		mov fs,dx
		mov gs,dx

; back to real mode
		and al,~1
		mov cr0,eax
	pop eax
	pop dx
	pop ds
	pop es
	pop fs
	pop gs
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:		vcpi_v86_to_real
; action:	uses VCPI to switch from V86 mode to pmode, then switches
;		to real mode
; inputs:	VCPI must be present
; outputs:	interrupts disabled, CY=1 if error, CY=0 if not
; minimum CPU:	386+
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

vcpi_v86_to_real:
	push esi
	push edi
	push cx
	push ebx
	push eax
; save CS value for return to real mode
		mov [_real_mode_cs],cs

; virtual-to-physical conversion value = code segment base address
		xor ebx,ebx
		mov bx,cs
		shl ebx,4

; set base addresses of the 16- and 32-bit code segment descriptors
		mov eax,ebx
		mov [_gdt2 + 2],ax
		mov [_gdt4 + 2],ax
		shr eax,16
		mov [_gdt2 + 4],al
		mov [_gdt4 + 4],al
		mov [_gdt2 + 7],ah
		mov [_gdt4 + 7],ah

; virtual-to-physical conversion value = data segment base address
		xor ebx,ebx
		mov bx,ds
		shl ebx,4

; set base addresses of the 16- and 32-bit data segment descriptors
		mov eax,ebx
		mov [_gdt3 + 2],ax
		mov [_gdt5 + 2],ax
		shr eax,16
		mov [_gdt3 + 4],al
		mov [_gdt5 + 4],al
		mov [_gdt3 + 7],ah
		mov [_gdt5 + 7],ah

; point the TSS descriptor to the LINEAR address of '_tss'
; (because paging is set up for identity-mapping, LINEAR=PHYSICAL)
		lea eax,[ebx + _tss]
		mov [_gdt6 + 2],ax
		shr eax,16
		mov [_gdt6 + 4],al
		mov [_gdt6 + 7],ah

; point the interrupt gates in the Interrupt Descriptor Table (IDT)
; to the default interrupt handler 'unhand'
		mov di,_idt
		mov cx,(_idt_end - _idt) / 8 ; number of IDT entries (gates)
do_idt:
		mov eax,unhand
		mov [di],ax		; set bits 15:0 of gate offset
		shr eax,16
		mov [di+6],ax		; set bits 31:16 of gate offset
		add di,byte 8		; next gate
		loop do_idt

; point '_gdt_ptr' to LINEAR/PHYSICAL address of the GDT;
; '_idt_ptr' to LINEAR/PHYSICAL address of the IDT
		lea eax,[ebx + _gdt]
		mov [_gdt_ptr + 2],eax
		lea eax,[ebx + _idt]
		mov [_idt_ptr + 2],eax

; DOS doesn't load .COM or .EXE files on a page (4K) boundary,
; so we must now find a page boundary for the page directory
		xor esi,esi
		mov si,_page_tables	; SI = offset of _page_tables
		add esi,ebx		; ESI = PHYSICAL address of _page_tables
		add esi,4095		; round to 4K boundary
		and esi,0xFFFFF000	; ESI = PHYSICAL address of page directory
		mov edi,esi
		sub edi,ebx		; DI = offset of page directory

; save page directory address for later use by VCPI
		mov [_vcpi_cr3],esi

; EAX = PHYSICAL address of page table
; (4K above the page directory)
		lea eax,[esi + 4096]
; create entry in page dir for page table
		or al,7			; Ring 3, writable, present
		mov [di + 0],eax

; advance DI from page directory to page table
		add di,4096

; get page table and partial GDT from the VCPI server
		mov si,_gdt7
		mov ax,0xDE01
		push ebx
			int 0x67
			mov [_vcpi_pmode_entry],ebx
		pop ebx
		cmp ah,0
		stc
		jne vcpi_error

; set up the VCPI control block for the switch to pmode
		lea eax,[ebx + _gdt_ptr]
		mov [_vcpi_gdtr],eax
		lea eax,[ebx + _idt_ptr]
		mov [_vcpi_idtr],eax
		lea esi,[ebx + _vcpi_control_block]
		mov ax,0xDE0C

		mov [_real_mode_sp],sp ; crock

; if all goes well, the interrupt will return at 'from_vcpi'
; (below, in the 32-bit code section)
		int 0x67

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

real2:

; back-to-real-mode sequence; based on 14.5 of 386INTEL.TXT (...continued)
; 4. load segment registers with values "appropriate to real mode"
; (byte-granular limit 64K-1, expand-up, writable, present)

mov byte [fs:dword 0xB8008],'4'

		mov ax,DATA16_SEL
		mov ss,ax
		mov ds,ax
		mov es,ax
; leave FS and GS with 32-bit limit ("unreal" mode)
; (using FS: to write debug characters to video memory)
;		mov fs,ax
;		mov gs,ax

; 5. clear the PE bit
mov byte [fs:dword 0xB800A],'5'
		mov eax,cr0
		and al,0xFE
		mov cr0,eax

; 6. far jump to true real mode
mov byte [fs:dword 0xB800C],'6'
		jmp far [_real_mode_ip]

; 7. restore real-mode segment registers
real3:
mov byte [fs:dword 0xB800E],'7'
		mov ax,cs
		mov ss,ax
		mov ds,ax
		mov es,ax

		xor ax,ax
		mov fs,ax
		mov gs,ax

; 8. load an IDT that is compatible with real-mode IVT
mov byte [fs:dword 0xB8010],'8'
		o32 lidt [_real_idt_ptr]
		clc
mov byte [fs:dword 0xB8012],'9'
		mov sp,[_real_mode_sp] ; crock
vcpi_error:
	pop eax
	pop ebx
	pop cx
	pop edi
	pop esi
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; 32-BIT CODE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	BITS 32

unhand:
	mov ax,LINEAR_SEL
	mov fs,ax
	mov byte [fs:0xB8000],'!'
	jmp short $

from_vcpi:
	cli

	mov ax,LINEAR_SEL
	mov fs,ax
	mov gs,ax
mov byte [fs:dword 0xB8000],'0'

	mov ax,DATA32_SEL
	mov ss,ax
	mov ds,ax
	mov es,ax

; back-to-real-mode sequence; based on 14.5 of 386INTEL.TXT
; 1. disable interrupts
;	(done above)

; 2. turn off paging
; 2a. jump to code memory which is identity-mapped
	; (already done)

; 2b. clear the PG bit
mov byte [fs:dword 0xB8002],'1'
	mov eax,cr0
	and eax,0x7FFFFFFF
	mov cr0,eax

; 2c. "Move zeros to CR3 to clear out the paging cache."
mov byte [fs:dword 0xB8004],'2'
	xor eax,eax
	mov cr3,eax

; 3. jump to 16-bit code segment (real2 is above, in the BITS 16 section)
mov byte [fs:dword 0xB8006],'3'
	jmp CODE16_SEL:real2

	BITS 16

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; name:		unboot
; action:	reload IVT, copy already-loaded bootsector to 0x7C00,
;		jump to it
; inputs:	arguments on stack per C function prototype
; outputs:	returns only if VCPI error occurs
; minimum CPU:	386+
; C prototype:	void unboot(const void *ivt_buf, unsigned ivt_len,
;			const void *boot_buf, unsigned boot_drive);
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

GLOBAL _unboot
_unboot:

; save registers
	push bp
		mov bp,sp

; make sure INT 67h is hooked before trying to call it
		xor ax,ax
		mov es,ax
		mov ax,[es:0x67*4 + 0]
		or  ax,[es:0x67*4 + 2]
		push ds
		pop es
		je no_vcpi

; check for VCPI
		mov ax,0xDE00
		int 0x67
		cmp ah,0
		jne no_vcpi

		call vcpi_v86_to_real
		jnc fix_ivt
; VCPI error:
	pop bp
	ret

; disable unreal mode
no_vcpi:
		xor ax,ax
		call unreal
fix_ivt:
mov ax,0xB800
mov fs,ax
mov byte [fs:dword 0x14],'A'
; copy IVT to address 0000:0000
		cli		; disable interrupts
		cld		; rep movsb auto-increments
		xor di,di
		mov es,di
		mov si,[bp + 4]	; ivt_buf
		mov cx,[bp + 6]	; ivt_len
		rep movsb

; copy bootsector or MBR to address 0000:7C00
		mov di,0x7C00
		mov si,[bp + 8]	; boot_buf
		mov cx,512	; one sector
		rep movsb

; zero the top 16 bits of the 32-bit registers
		xor eax,eax
		xor ebx,ebx
		xor ecx,ecx
		xor edx,edx
		xor esi,esi
		xor edi,edi

; set DL to boot drive
		mov dl,[bp + 10]

		xor ebp,ebp

; zero segment registers
		mov ss,bx
		mov ds,bx
		mov es,bx
		mov fs,bx
		mov gs,bx

; set stack to 7C00h (and zero top 16 bits of ESP)
		mov esp,0x7C00

; jump to 0000:7C00h (zeroes CS)
		jmp 0:0x7C00

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; DATA
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

SEGMENT _DATA PUBLIC CLASS=DATA ALIGN=2

_real_mode_ip:
	dw real3
_real_mode_cs:
	dw 0

_real_mode_sp:
	dw 0

_gdt:
; null descriptor
	dw 0			; limit 15:0
	dw 0			; base 15:0
	db 0			; base 23:16
	db 0			; type
	db 0			; limit 19:16, flags
	db 0			; base 31:24

; data descriptor
LINEAR_SEL	equ	$-_gdt
_gdt1:
	dw 0xFFFF		; limit 64K
	dw 0			; base address
	db 0
	db 0x92			; present, ring 0, data, expand-up, writable
; use "db 0xCF" here to ENABLE unreal mode
; (page-granular 20-bit segment limit = 4G-1)
	db 0xCF			; 0 = byte-granular 16-bit limit = 64K-1
	db 0
CODE32_SEL equ $-_gdt
_gdt2:
	dw 0xFFFF
	dw 0
	db 0
	db 0x9B		; present, ring 0, code, non-conforming, readable, accessed
	db 0xCF		; page-granular, 32-bit
	db 0
DATA32_SEL equ $-_gdt
_gdt3:
	dw 0xFFFF
	dw 0
	db 0
	db 0x93		; present, ring 0, data, expand-up, writable, accessed
	db 0xCF		; page-granular, 32-bit
	db 0
CODE16_SEL equ $-_gdt
_gdt4:
	dw 0xFFFF
	dw 0
	db 0
	db 0x9B		; present, ring 0, code, non-conforming, readable, accessed
	db 0		; byte-granular, 16-bit
	db 0
DATA16_SEL equ $-_gdt
_gdt5:
	dw 0xFFFF
	dw 0
	db 0
	db 0x93		; present, ring 0, data, expand-up, writable, accessed
	db 0		; byte-granular, 16-bit
	db 0
TSS_SEL equ $-_gdt
_gdt6:
	dw 103
	dw 0
	db 0
	db 0x89		; present, ring 0, available 32-bit TSS
	db 0
	db 0
VCPI_SEL equ $-_gdt
_gdt7:
	dd 0, 0		; code and two scratch descriptors used by VCPI

	dd 0, 0

	dd 0, 0
_gdt_end:

_idt:
	%rep 32

	dw 0		; low 16 bits of ISR address
	dw CODE32_SEL	; selector
	db 0		; word count
	db 0x8E		; access byte: Present, Ring 0, '386 intr gate
	dw 0		; high 16 bits of ISR

	%endrep
_idt_end:

; to save 6 bytes of memory, this structure can be stored at (gdt+0)
; in the NULL descriptor, but that can be confusing
_gdt_ptr:
	dw _gdt_end - _gdt - 1	; GDT limit
	dd _gdt			; linear address of GDT; gets patched above

_idt_ptr:
	dw _idt_end - _idt - 1	; IDT limit
	dd _idt			; linear, physical address of IDT

_real_idt_ptr:
	dw 1023			; limit
	dd 0			; IDT (IVT, actually) at address 0

_vcpi_pmode_entry:
	dd 0
	dw VCPI_SEL

_vcpi_control_block:
_vcpi_cr3:
	dd 0		; initialized above
_vcpi_gdtr:
	dd _gdt_ptr	; fixed up above
_vcpi_idtr:
	dd _idt_ptr	; fixed up above
_vcpi_ldtr:
	dw 0
_vcpi_tr:
	dw TSS_SEL
_vcpi_eip:
	dd from_vcpi	; no fixup for this
_vcpi_cs:
	dw CODE32_SEL

_tss:
	dw 0, 0		; back link

	dd 0		; ESP0
	dw DATA32_SEL, 0; SS0, reserved

	dd 0		; ESP1
	dw 0, 0		; SS1, reserved

	dd 0		; ESP2
	dw 0, 0		; SS2, reserved

	dd 0		; CR3
	dd 0, 0		; EIP, EFLAGS
	dd 0, 0, 0, 0	; EAX, ECX, EDX, EBX
	dd 0, 0, 0, 0	; ESP, EBP, ESI, EDI
	dw 0, 0		; ES, reserved
	dw 0, 0		; CS, reserved
	dw 0, 0		; SS, reserved
	dw 0, 0		; DS, reserved
	dw 0, 0		; FS, reserved
	dw 0, 0		; GS, reserved
	dw 0, 0		; LDT, reserved
	dw 0, 0		; debug, I/O permission bitmap (none)

_page_tables:
	times 4096 db 0         ; padding to 4K boundary
	times 4096 db 0         ; page directory somewhere in here
	times 4096 db 0         ; page table for bottom 4 meg
