/*----------------------------------------------------------------------------
Fast reboot from DOS
Chris Giese	<geezer@execpc.com>	http://my.execpc.com/~geezer/
This code is public domain (no copyright).
You can do whatever you want with it.

27 Jul. 2015:
- Call to unreal_mode() deleted; that functionality is now incorporated
  into unboot()
- Added VCPI code to switch from V86 mode to pmode, then to real mode.
  It ALMOST works -- the system loads the MBR, displays a boot menu
  with timeout, starts loading DOS 7.1, loads EMM386.EXE...then hangs.
  If I load EMM386.EXE, then comment it out in CONFIG.SYS before
  rebooting, the PC still hangs when I reboot.

25 Jul. 2015:
- Added code to abort if Windows is running.
- All of the assembly-language code at the bottom of main() has been
  moved to a function named "unboot" in the assembly-language file.
- The assembly-language file has been renamed to "UNBOOTA.ASM".
  The C code has been renamed to "UNBOOTC.C".
- It now builds and runs with 16-bit Watcom C.

23 Jun. 2015:
- MEMDUMP.ASM renamed to DUMP-IVT.ASM and changed thus:
- It now dumps only 1024 bytes (IVT); not 1280 (IVT + BDA)
- It now wipes the floppy disk automatically
- It now creates a 1024-byte output file ("IVT.BIN") automatically

Feb. 2015:
- Initial release

I originally wanted to load and run NTLDR from DOS. Rather than do that,
I decided to generalize the problem by loading and running a bootsector
or MBR.

Problems:
- Must restore interrupt vectors to their pre-DOS values. There is
  no easy way to do this. MS-DOS saves some of the old handler
  addresses at 0070h:0100h but not all of them. I wrote a bootsector
  (DUMP-IVT.ASM) that writes out the vectors to a file on the floppy
  disk. You must create a floppy disk containing this bootsector,
  run it on the target system, then make the file that it creates
  (IVT.BIN) available to the UNBOOT program.

- Reboot fails after running Turbo C++ 3.0; which is a 16-bit DPMI
  program. I think it leaves invalid values in the segment descriptor
  caches; because either enabling or disabling unreal mode fixes it.

- Enabling or disabling unreal mode entails a temporary switch to
  pmode; which is not permitted by EMM386.EXE. Must use VCPI to
  disable EMM386.EXE. I have a good idea of how to do this but I'm
  still working on it.
----------------------------------------------------------------------------*/
/* the VCPI code now in UNBOOTA.ASM assumes tiny memory model: */
#if !defined(__TINY__)
#error I want TINY memory model!
#endif

#include <stdlib.h> /* malloc(), free() */
#include <string.h> /* memcpy(), _fmemcpy() */
#include <stdio.h> /* FILE, fopen(), fread(), fclose(), printf() */
#include <bios.h> /* _DISK_..., biosdisk() */
#include <dos.h> /* FP_SEG(), MK_FP(), disable(), enable() */
#if 0
#include <stdint.h>
#else
typedef unsigned char	uint8_t;
typedef unsigned short	uint16_t;
typedef unsigned long	uint32_t;
#endif

#if defined(__WATCOMC__)
#define	peekb(S,O)	*(uint8_t  far *)MK_FP(S,O)
#define	peek(S,O)	*(uint16_t far *)MK_FP(S,O)
#endif
/****************************************************************************/
static unsigned read_le16(const void *buf0)
{
	const unsigned char *buf = buf0;
	unsigned rv;

	rv  = buf[1];
	rv <<= 8;
	rv |= buf[0];
	return rv;
}
/****************************************************************************/
static unsigned long read_le32(const void *buf0)
{
	const unsigned char *buf = buf0;
	unsigned long rv;

	rv = buf[3];
	rv <<= 8;
	rv |= buf[2];
	rv <<= 8;
	rv |= buf[1];
	rv <<= 8;
	rv |= buf[0];
	return rv;
}
/****************************************************************************/
static int is_fat_bootsector(const void *buf0)
{
	const unsigned char *buf = buf0;
	int good = 1;
	unsigned i;

	if(buf0	== NULL)
		return 0;
/* must start with 16-bit JMP or 8-bit JMP plus NOP */
	if(buf[0] == 0xE9)
		/* OK */;
	else if(buf[0] == 0xEB && buf[2] == 0x90)
		/* OK */;
	else
		good = 0;
/* bytes per sector must be a power of 2 */
	i = read_le16(&buf[11]);
	if(i ==	0 || ((i - 1) &	i) != 0)
		good = 0;
/* sectors per cluster must be a power of 2 */
	i = buf[13];
	if(i ==	0 || ((i - 1) &	i) != 0)
		good = 0;
/* NEW: must be at least one reserved (boot) sector */
	if(read_le16(&buf[14]) == 0)
		good = 0;
/* NEW: must be at least one FAT */
	if(buf[16] == 0)
		good = 0;
/* NEW: 16-bit and 32-bit sector count can't both be zero */
	if(read_le16(&buf[19]) == 0 &&	read_le32(&buf[32]) == 0L)
		good = 0;
/* must have sensible CHS geometry */
	i = read_le16(&buf[24]);
	if(i ==	0 || i > 63)
		good = 0;
	i = read_le16(&buf[26]);
	if(i ==	0 || i > 255)
		good = 0;
/* bootsector magic values
fat_mount() is using only a 48-byte buffer, so skip this test:
	if(buf[510] != 0x55 || buf[511] != 0xAA)
		good = 0; */
	return good;
}
/*----------------------------------------------------------------------------
READ FROM DISK USING BIOS FUNCTIONS (INT 13h)
----------------------------------------------------------------------------*/
typedef struct
{
	enum
	{
		UL_UNINIT = 0, UL_CHS, UL_LBA
	} use_lba;
	unsigned char int13_drive_num;	/* 0=floppy, 0x80=hard disk */
	unsigned char heads, spt;	/* valid if use_lba == UL_CHS */
	unsigned long num_sects;
	unsigned bps;			/* bytes per sector */
} disk_t;
/****************************************************************************/
static int read_disk_chs(unsigned drive, void *buf, unsigned c,
		unsigned h, unsigned s, unsigned num_sects)
{
	unsigned tries,	err;
	struct SREGS sregs;
	union REGS regs;

	if(num_sects == 0)
		return 0;
//	if((buf == NULL && cmd != _DISK_RESET) || h > 255
	if(buf == NULL || h > 255 || s == 0 || s > 63 || num_sects > 255)
	{
		printf("Error in read_disk_chs(): bad arg\n");
		return -1;
	}
/* set up registers for INT 13h AH=02/03h */
	regs.h.dh = h;	regs.h.dl = drive;
	regs.h.ch = c;	regs.h.cl = (((c >> 2) & 0xC0) | s);
	sregs.es = FP_SEG(buf);
	regs.x.bx = FP_OFF(buf);
/* make 3 attempts */
	for(tries = 3; tries !=	0; tries--)
	{
/*		if(cmd == _DISK_RESET)
			regs.h.ah = 0x00;
		else if(cmd == _DISK_STATUS)
			regs.h.ah = 0x01;
		else if(cmd == _DISK_READ) */
			regs.h.ah = 0x02;
/*		else if(cmd == _DISK_WRITE)
			regs.h.ah = 0x03;
		else if(cmd == _DISK_VERIFY)
			regs.h.ah = 0x04;
		else if(cmd == _DISK_FORMAT)
			regs.h.ah = 0x05;
		else
			goto ERR; invalid command */
		regs.h.al = num_sects;
		int86x(0x13, &regs, &regs, &sregs);
		if(regs.x.cflag == 0)
			return 0;
		err = regs.h.ah;
/* reset disk */
		regs.h.ah = 0;
		int86(0x13, &regs, &regs);
	}
	printf("Error in read_disk_chs(): INT 0x13 error 0x%02X\n", err);
	return -1;
}
/****************************************************************************/
static int read_disk_lba(unsigned drive, void *buf, unsigned long lba,
		unsigned num_sects)
{
/* INT 13h AH=42h/AH=43h command packet: */
#pragma	pack(1)
	struct
	{
		uint8_t	 pkt_len;
		uint8_t	 res;
		uint8_t	 num_sects;
		uint8_t	 res1;
		uint16_t buf_off;
		uint16_t buf_seg;
		uint32_t lba31_0;
		uint32_t lba63_32;
	} lba_cmd_pkt;
	unsigned tries,	err;
	struct SREGS sregs;
	union REGS regs;

	if(num_sects == 0)
		return 0;
	if(buf == NULL || num_sects > 255)
	{
		printf("Error in read_disk_lba(): bad arg\n");
		return -1;
	}
/* fill out the INT 13h AH=4xh command packet */
	memset(&lba_cmd_pkt, 0,	sizeof(lba_cmd_pkt));
	lba_cmd_pkt.pkt_len = sizeof(lba_cmd_pkt);
	lba_cmd_pkt.num_sects = num_sects;
	lba_cmd_pkt.buf_off = FP_OFF(buf);
	lba_cmd_pkt.buf_seg = FP_SEG(buf);
	lba_cmd_pkt.lba31_0 = lba;
/* set up registers for INT 13h AH=4xh */
	sregs.ds = FP_SEG(&lba_cmd_pkt);
	regs.x.si = FP_OFF(&lba_cmd_pkt);
	regs.h.dl = drive;
/* make 3 attempts */
	for(tries = 3; tries !=	0; tries--)
	{
/*		if(cmd == _DISK_RESET)
			regs.h.ah = 0x00;
		else if(cmd == _DISK_STATUS)
			regs.h.ah = 0x01;
		else if(cmd == _DISK_READ) */
			regs.h.ah = 0x42;
/*		else if(cmd == _DISK_WRITE)
			regs.h.ah = 0x43;
		else if(cmd == _DISK_VERIFY)
			regs.h.ah = 0x44;
		else
			goto ERR; invalid command */
		int86x(0x13, &regs, &regs, &sregs);
		if(regs.x.cflag == 0)
			return 0;
		err = regs.h.ah;
/* reset disk */
		regs.h.ah = 0;
		int86(0x13, &regs, &regs);
	}
	printf("Error in read_disk_lba(): INT 0x13 error 0x%02X\n", err);
	return -1;
}
/****************************************************************************/
static int init_fd(disk_t *d)
{
	char buf[512];
	int i;

	d->use_lba = UL_CHS;
/* read sector 0 */
	if((i = read_disk_chs(d->int13_drive_num, buf, 0, 0, 1, 1)) != 0)
		return i;
/* test if FAT bootsector */
	if(is_fat_bootsector(buf))
	{
		unsigned long ns;

/* get CHS from FAT bootsector */
		d->spt		= read_le16(&buf[24]);
		d->bps		= read_le16(&buf[11]);
		d->heads	= read_le16(&buf[26]);
		ns		= read_le16(&buf[19]);
		if(ns == 0)
			ns	= read_le32(&buf[32]);
		d->num_sects = ns;
		printf("init_fd(): floppy drive 0x%02X using "
			"CHS=%u:%u:%u (from FAT bootsector)\n",
			d->int13_drive_num, (unsigned)(ns / d->heads
			/ d->spt), d->heads,	d->spt);
	}
	else
	{
		unsigned long num_tracks;

/* probe sectors-per-track value */
		printf("init_fd(): probing floppy geometry, "
			"please wait...\n");
		for(i =	1; i < 40; i++)
		{
			if(read_disk_chs(d->int13_drive_num, buf, 0, 0, i, 1))
				break;
		}
		d->spt		= i - 1;
/* other values are assumed: */
		d->bps		= 512;
		d->heads	= 2;
		num_tracks	= 80;
		d->num_sects = num_tracks * d->heads *	d->spt;
		printf("init_fd(): floppy drive 0x%02X using "
			"CHS=%lu:%u:%u (probed)\n", d->int13_drive_num,
			num_tracks, d->heads, d->spt);
	}
	return 0;
}
/****************************************************************************/
static int init_hd(disk_t *d)
{
	unsigned j, max_cyl_minus1;
	unsigned char buf[26];
	struct SREGS sregs;
	union REGS regs;

/* get hard drive CHS geometry from BIOS */
	regs.h.ah = 0x08;
	regs.h.dl = d->int13_drive_num;
	int86(0x13, &regs, &regs);
	if(regs.x.cflag)
	{
		printf("Error in init_hd(): can't get "
			"geometry for hard drive 0x%02X\n",
			d->int13_drive_num);
		return -1;
	}
	d->heads = regs.h.dh + 1;
	d->spt = regs.h.cl & 0x3F;
	j = regs.x.cx;
	j &= 0xC0;
	j <<= 2;
	max_cyl_minus1 = j | regs.h.ch;
/* use LBA if drive and BIOS support it */
	regs.h.ah = 0x41;
	regs.x.bx = 0x55AA;
	regs.h.dl = d->int13_drive_num;
	int86(0x13, &regs, &regs);
	if(regs.x.cflag == 0 && regs.x.bx == 0xAA55)
	{
		buf[0] = sizeof(buf);
		sregs.ds = FP_SEG(buf);
		regs.x.si = FP_OFF(buf);
		regs.h.dl = d->int13_drive_num;
		regs.h.ah = 0x48;
		int86x(0x13, &regs, &regs, &sregs);
		if(regs.x.cflag == 0)
		{
			d->use_lba = UL_LBA;
/* it's actually a 64-bit value: */
			d->num_sects = read_le32(&buf[16]);
			d->bps = read_le16(&buf[24]);
			printf("init_hd(): hard drive 0x%02X "
				"using LBA\n", d->int13_drive_num);
			return 0;
		}
	}
/* else use CHS */
	printf("init_hd(): hard drive 0x%02X using CHS=?:%u:%u "
		"(from INT 13h AH=08h)\n", d->int13_drive_num, d->heads,
		d->spt);
	d->use_lba = UL_CHS;
	d->num_sects = (max_cyl_minus1 + 2L) * d->heads * d->spt;
	d->bps = 512;
	return 0;
}
/****************************************************************************/
static int init_disk(disk_t *d)
{
	unsigned char i, num_drives;

	if(d->int13_drive_num < 0x80)
	{
		if((i = peek(0x40, 0x10)) & 1)
			num_drives = ((i >> 6) & 0x03) + 1;
		else
			num_drives = 0;
		printf("%u floppy disk(s) per BIOS\n", num_drives);
	}
	else
	{
		num_drives = peekb(0x40, 0x75);
		printf("%u hard disk(s) per BIOS\n", num_drives);
	}
	if((d->int13_drive_num & 0x7F) >= num_drives)
	{
		printf("Error in init_disk(): disk 0x%02X does not exist\n",
			d->int13_drive_num);
		return -1;
	}
	return (d->int13_drive_num < 0x80)
		? init_fd(d) : init_hd(d);
}
/****************************************************************************/
static int read_disk(disk_t *d, void *buf, unsigned long lba,
		unsigned num_sects)
{
	unsigned s, h, c;
	int i;

/*	if(d->use_lba == UL_UNINIT) */
	if(d->use_lba != UL_LBA && d->use_lba != UL_CHS)
	{
		if((i = init_disk(d)) != 0)
			return i;
	}
	if(d->use_lba == UL_LBA)
		return read_disk_lba(d->int13_drive_num, buf,
			lba, num_sects);
	/* else if(d->use_lba == UL_CHS)*/
/* convert LBA to CHS */
	s = (unsigned)(lba % d->spt) + 1;
	h = (unsigned)((lba / d->spt) % d->heads);
	c = (unsigned)((lba / d->spt) / d->heads);
	return read_disk_chs(d->int13_drive_num, buf,
		h, c, s, num_sects);
}
/*----------------------------------------------------------------------------
UNBOOT
----------------------------------------------------------------------------*/
/* actual length of the IVT is 1024 bytes, but based on testing I've done,
we need only restore the bottom 512 bytes (see notes below) */
#define	IVT_LEN	512

#pragma pack(1)
typedef struct
{
	uint16_t off, seg;
} vector_t;

/* IMPORTS
from UNBOOTA.ASM: */
void _cdecl unboot(const void *ivt_buf, unsigned ivt_len,
		const void *boot_buf, unsigned boot_drive);
/****************************************************************************/
int main(void)
{
	static unsigned char ivt_buf[IVT_LEN], boot_buf[512];
	static const char *img_name = "ivt.bin";
/* part_num: -1=boot from MBR. Partition numbers on my hard disk:
	0=FAT16 C:   1=NTFS   2=Linux ext2   3=FAT32 D:
Only the MBR and my C: have valid boot code in sector 0 */
	static const int part_num = -1;
/**/
	disk_t d = { 0, 0x80 }; /* drive: 0=floppy, 0x80=hard disk */
	unsigned long lba;
	unsigned i, error;
	union REGS regs;
	vector_t *v;
	FILE *f;

/* check if Windows running */
	regs.x.ax = 0x1600;
	int86(0x2F, &regs, &regs);
	if(regs.h.al != 0x00 && regs.h.al != 0x80)
	{
		printf("Sorry, can't run this code from Windows\n");
		return 1;
	}
/* open and load the memory image file made by DUMP-IVT.ASM */
	if((f = fopen(img_name, "rb")) == NULL)
	{
		printf("Error: can't open input file '%s'\n", img_name);
		return 2;
	}
	i = fread(ivt_buf, 1, IVT_LEN, f);
	fclose(f);
	if(i < IVT_LEN)
	{
		printf("Error: wanted %u bytes from file %s, "
			"got only %u\n", IVT_LEN, img_name, i);
		return 3;
	}
/* validate interrupt vector table (IVT) */
	v = (vector_t *)ivt_buf;
	error = 0;
	for(i = 0; i < IVT_LEN / 4; i++)
	{
/* vector = 0000:0000 */
		if(v[i].off == 0 && v[i].seg == 0)
			/* OK */;
/* vector points to ROM BIOS (video or motherboard) */
		else if(v[i].seg >= 0xC000)
			/* OK */;
		else
		{
			error = 1;
			printf("Error: interrupt vector %u (0x%02X) may"
				" be invalid (-> %04X:%04X)\n", i, i,
				v[i].seg, v[i].off);
		}
	}
	if(error)
	{
		printf("Error: invalid interrupt table in memory dump\n");
		return 4;
	}
	lba = 0L;
/* load MBR */
	if(d.int13_drive_num >= 0x80 && part_num >= 0)
	{
		unsigned char *pte;

		if(read_disk(&d, boot_buf, lba, 1) != 0)
ERR:		{
			printf("Error: can't read sector %lu of disk "
				"0x%02X\n", lba, d.int13_drive_num);
			return 5;
		}
		pte = &boot_buf[446 + 16 * part_num];
		lba = read_le32(&pte[8]);
		if(pte[4] == 0 || lba == 0)
		{
			printf("Error: partition %u on drive 0x%02X does "
				"not exist\n", part_num, d.int13_drive_num);
			return 6;
		}
	}
/* load bootsector (or MBR...) */
	if(read_disk(&d, boot_buf, lba, 1) != 0)
		goto ERR;
/* un-boot: */
	unboot(ivt_buf, IVT_LEN, boot_buf, d.int13_drive_num);
/* in normal operation, this code is never reached: */
	printf("VCPI error: INT 67h AX=DE01h failed\n");
	return 0;
}
